from asyncio.subprocess import DEVNULL
from asyncio.subprocess import PIPE
from asyncio.subprocess import Process
from asyncio.subprocess import STDOUT
from asyncio.subprocess import SubprocessStreamProtocol

_ = (DEVNULL, PIPE, STDOUT, Process, SubprocessStreamProtocol)
